<?php
/**
 * CREATE  TABLE IF NOT EXISTS `action_call` (
 * `id` INT NOT NULL AUTO_INCREMENT ,
 * `caller` VARCHAR(255) NOT NULL ,
 * `sessionID` VARCHAR(255) NOT NULL ,
 * `tabID` VARCHAR(255) NULL COMMENT 'NOT USED YET' ,
 * `called` TIMESTAMP NOT NULL ,
 * PRIMARY KEY (`id`) )
 * ENGINE = InnoDB
 *
 * The followings are the available columns in table 'action_call':
 * @property integer $id
 * @property string $caller
 * @property string $sessionID
 * @property string $tabID
 * @property string $called
 *
 * The followings are the available model relations:
 * @property CallData[] $callDatas
 * 
 * The followings are from getter functions:
 * @property array $answers
 * @property bool $completed
 * @property CallData|null $firstUnanswered
 */
class ActionCall extends CActiveRecord {
	/**
	 * @see CActiveRecord::init()
	 */
	public function init() {
		parent::init();
		$this->sessionID = $this->currentSessionID();
	}

    /**
     * For testing purposes only we use this help function
     * @return string
     */
    protected function currentSessionID() {
        $session = new CHttpSession();
        $sessionID = $session->sessionID;
        if( Hlp::isEmpty($sessionID) ) {  //case of phpunit testing where we are not on a real browser
            return 'SESSION_ID_ONE';    //totally dummy
        }
        return $sessionID;
    }

    /**
     * @static
     * @param Controller $controller
     * @param array[] $calls
     * @param array $params
     * @throws CException
     * @return array
     */
    public static function call($controller, $calls, $params=array()) {
        if($controller->action == null) {
            throw new CException('You tried to create a call from a controller with no active action');
        }

        $caller = CHtml::normalizeUrl( array($controller->route) + $params );
        $model = new ActionCall();
        $model->caller = $caller;

        $models = $model->getAllRelated($calls);
        unset($model);  //just to be sure

        //try to find a completed call
        $model = self::getFirst($models,true);
        //dd($models);
        if($model!=null) {
            $answers = $model->answers;
            $model->delete();
            return $answers;
        }

        //try to find an incomplete call
        $model = self::getFirst($models,false);
        if($model==null) { //case it doesn't exist (if all are completed then they are only leftovers and we dont care about them)
            $class = __CLASS__;
            $model = new $class();
            $model->caller = $caller;
            $model->createNewCall($calls);
        }
        //if($model->completed == true) { throw new CException('by now the action call model must have been an incomplete one'); }
        $callData = $model->firstUnanswered;    //PERFORM A CALL
        //if($callData == null) { throw new CException('if all are answered then we should have returned the data'); }
        $controller->redirect($callData->callee);
    }

    /**
     * TESTED
     * Checks if this call contains all the callees inside the calls configuration
     * @param array $calls
     * @return bool
     * It makes sense to call this function only on saved records which have callees
     */
    public function isRelated($calls) {
        $callDatas = $this->callDatas;
        if( count($callDatas) != count($calls) ) {
            return false;
        }

        $isRelated = true;
        foreach($callDatas as $callData) {
            $isRelated = $isRelated && $callData->isFoundInCalls($calls);
        }
        return $isRelated;
    }

    /**
     * TESTED
     * @param array $calls
     * @return ActionCall[]
     */
    public function getAllRelated($calls) {
        $models = self::model()->findAllByAttributes(array('caller' => $this->caller));
        $related = array();
        foreach($models as $model) {
            if( $model->isRelated($calls) ) {
                $related[] = $model;
            }
        }
        return $related;
    }

    /**
     * TESTED
     * Return the first call which is NOT completed,
     * or return null if all calls, in question, are completed
     * @param array|null $models
     * @param bool $completed
     * @return ActionCall|null
     */
    public static function getFirst($models=null,$completed=false) {
        if($models===null) {
            $models = self::model()->findAll();
        }

        foreach($models as $model) {
            if($model->completed == $completed) {
                return $model;
            }
        }
        return null;
    }

    /**
     * TESTED
     * @return array
     */
    public function getAnswers() {
        $answers = array();
        foreach($this->callDatas as $callData) {
            $answers[$callData->data_name] = $callData->data;
        }
        return $answers;
    }

    /**
     * TESTED
     * Get the first call of all the calls of this caller which is not answered
     * @return CallData|null
     */
    public function getFirstUnanswered() {
        foreach($this->callDatas as $callData) {
            if($callData->answered == false) {
                return $callData;
            }
        }
        return null;
    }

    /**
     * TESTED
     * @param array $calls
     * @throws CException
     */
    public function createNewCall($calls) {
        $this->called = TimeHelper::nowstamp();

        $transaction = Yii::app()->db->beginTransaction();
        try {
            $this->save(false);

            while( ( $callData=array_shift($calls) ) !== null ) {
                $model = new CallData();
                $model->callee = $callData['callee'];
                $model->data_name = $callData['data_name'];
                $model->answered = false;
                $model->action_call_id = $this->id;
                $model->save(false);
            }

            $transaction->commit();
        }
        catch(Exception $e) {
            $transaction->rollBack();
            throw new CException('Failed saving a new call');
        }
    }

    /**
     * TESTED
     * @return boolean
     */
    public function getCompleted() {
        $completed = true;
        foreach($this->callDatas as $callData) {
            $completed = $completed && $callData->answered;
        }
        return $completed;
    }

    /**
     * Returns the static model of the specified AR class.
     * @param string $className
     * @return ActionCall the static model class
     */
	public static function model($className=__CLASS__) {
		return parent::model($className);
	}

	/**
	 * @return string the associated database table name
	 */
	public function tableName() {
		return '{{action_call}}';
	}

	/**
	 * @return array validation rules for model attributes
	 */
	public function rules() {
		// NOTE: you should only define rules for those attributes that will receive user inputs.
		return array(
			array(
                array('caller','sessionID','called'),
                'required',
            ),
			array(
                array('caller','sessionID','tabID'),
                'length',
                'max'=>255,
            ),
            array(
                array('caller','sessionID'),
                'safe',
                'on'=>'search',
            ),
		);
	}

	/**
	 * @return array relational rules.
	 */
	public function relations() {
		return array(
			'callDatas' => array(self::HAS_MANY, 'CallData', 'action_call_id'),
		);
	}
	
	/**
     * REMEMBER default scope is applied also in relations
	 * Only records with the current sessionID will be considered
	 * @see CActiveRecord::defaultScope()
     * @return array
     */
	public function defaultScope() {
		return array(
            //'condition' => "sessionID = '{$this->currentSessionID()}'",
			'condition' => 'sessionID = :sessionID',
            'params' => array(
                ':sessionID' => $this->currentSessionID(),
            ),
		);
	}

	/**
	 * @return array customized attribute labels (name=>label)
	 */
	public function attributeLabels() {
		return array(
		);
	}

	/**
	 * Retrieves a list of models based on the current search/filter conditions.
	 * @return CActiveDataProvider the data provider that can return the models based on the search/filter conditions.
	 */
	public function search() {  // Warning: Please modify the following code to remove attributes that should not be searched.
		$criteria=new CDbCriteria;
		$criteria->compare('caller',$this->caller,true);
        $criteria->compare('sessionID',$this->sessionID,true);
		return new CActiveDataProvider($this, compact('criteria'));
	}
}